/*
 * Author:  František Dvořák <valtri@users.sourceforge.net>
 * License: public domain
 *
 * $Id: statistics.c.c $
 */

#include <stddef.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

#include "statistics.h"

#define MY_INFTY 1e20


/*
 * vrátí základní info o daném rozdělení
 */
void stat_expected(size_t n, double *x, double *p, double *avg, double *aber) {
	double sum, sum2;
	size_t i;

	sum = 0;
	sum2 = 0;
	for (i = 0; i < n; i++) {
		sum += x[i]*p[i];
		sum2 += x[i]*x[i]*p[i];
	}

	*avg = sum;
	*aber = sqrt(sum2 - sum*sum);
}


/*
 * vrátí info z výsledků
 */
void stat_real(size_t n, double *y, double *avg, double *aber) {
	double sum, sum2;
	size_t i;

	sum = 0;
	sum2 = 0;
	for (i = 0; i < n; i++) {
		sum += y[i];
		sum2 += y[i]*y[i];
	}

	*avg = sum / n;
	*aber = sqrt(sum2 / n - *avg * *avg);
}


/*
 * nalezení nejmenší dvojice intervalů
 */
static size_t intervals_get_min(interval_t *intervals, int n) {
	int index, i;
	double min, len;

	index = 0;
	min = intervals[1].x2 - intervals[0].x1;
	for (i = 1; i + 2 < n; i++) {
		len = intervals[i+1].x2 - intervals[i].x1;
		if (len < min) {
			min = len;
			index = i;
		}
	}

	return index;
}


/*
 * sloučení dvou intervalů
 */
static void intervals_merge(interval_t *intervals, int n, int index) {
	int i;

	intervals[index].x2 = intervals[index + 1].x2;
	for (i = index + 1; i < n; i++) memcpy(&intervals[i], &intervals[i+1], sizeof(interval_t));
}


/*
 * nalezení intervalu, který obsahuje číslo 'x'
 */
static int intervals_get_interval(interval_t *intervals, int n, double x) {
	size_t i;

	if (x < intervals[0].x1) return 0;
	for (i = 0; i < (size_t)n - 2; i++)
		if (x >= intervals[i].x1 && x < intervals[i].x2) return i + 1;
	return n - 1;
}


/*
 * příprava na vypsání histogramu (zpracuje se x-ová osa - intervaly)
 */
void histogram_init(histogram_t *histogram, double *x, size_t n_p, int width, int height) {
	int index;
	size_t i, n_i;
	void *tmp;

	histogram->height = height;
	histogram->width = width;

	n_i = n_p - 1;
	histogram->intervals = malloc(n_i * sizeof(interval_t));
	histogram->counters = malloc(width * sizeof(int));
	memset(histogram->counters, 0, histogram->width * sizeof(int));

	for (i = 0; i < n_i; i++) {
		histogram->intervals[i].x1 = (x[i] + x[i + 1]) / 2;
		histogram->intervals[i].x2 = (x[i + 1] + x[i + 2]) / 2;
	}

	while (n_i + 1 > (size_t)width) {
		index = intervals_get_min(histogram->intervals, n_i);
		intervals_merge(histogram->intervals, n_i, index);
		n_i--;
	}

	tmp = realloc(histogram->intervals, n_i * sizeof(interval_t));
	if (tmp) histogram->intervals = tmp;

	if ((size_t)histogram->width > n_i + 1) histogram->width = n_i + 1;
}


/*
 * připraví histogram očekávaných hodnot
 */
void histogram_prep_expected(histogram_t *histogram, double *x, double *p, size_t n_p) {
	double *counters, max;
	int index;
	int k;

	counters = malloc(n_p * sizeof(double));
	memset(counters, 0, n_p * sizeof(double));

	/* určení četností v intervalech */
	for (k = 0; (size_t)k < n_p; k++) {
		index = intervals_get_interval(histogram->intervals, histogram->width, x[k]);
		counters[index] += p[k];
	}

	/* přepočet na "ASCII jednotky" */
	max = histogram->counters[0];
	for (k = 1; k < histogram->width; k++)
		if (counters[k] > max)
			max = counters[k];
	for (k = 0; k < histogram->width; k++) {
		histogram->counters[k] = (int)(counters[k] / max * histogram->height + 0.5);
	}

	free(counters);
}


/*
 * připraví histogram experimentálních hodnot
 */
void histogram_prep_output(histogram_t *histogram, double *y, size_t n_y) {
	int *counters = histogram->counters;
	int k;
	int index, max;
	
	/* určení četností v intervalech */
	memset(counters, 0, histogram->width * sizeof(int));
	for (k = 0; (size_t)k < n_y; k++) {
		index = intervals_get_interval(histogram->intervals, histogram->width, y[k]);
		counters[index]++;
	}

	/* přepočet na "ASCII jednotky" */
	max = counters[0];
	for (k = 1; k < histogram->width; k++)
		if (counters[k] > max)
			max = counters[k];
	for (k = 0; k < histogram->width; k++)
	  counters[k] = (int)((double)counters[k] / max * histogram->height + 0.5);
}


/*
 * vypsání histogramu
 */
void histogram_render(histogram_t *histogram) {
	int i, j;

	/* ASCII art renderování :-) */
	for (j = histogram->height; j > 0; j--) {
		printf("|");
		for (i = 0; i < histogram->width; i++) {
			if (histogram->counters[i] > j - 1) printf("*");
			else printf(" ");
		}
		printf("\n");
	}
	printf("+");
	for (i = 0; i < histogram->width; i++) printf("-");
	printf("\n\n");
}


/*
 * uvolní připravený histogram z paměti
 */
void histogram_free(histogram_t *histogram) {
	if (histogram->counters) {
		free(histogram->counters);
		histogram->counters = NULL;
	}
	if (histogram->intervals) {
		free(histogram->intervals);
		histogram->intervals = NULL;
	}
}
